#   ==============================================================================
#   Copyright (c) 2024 Botts Innovative Research, Inc.
#   Date:  2024/6/26
#   Author:  Ian Patterson
#   Contact Email:  ian@botts-inc.com
#   ==============================================================================
#
#  Author: Ian Patterson <ian@botts-inc.com>
#
#  Contact Email: ian@botts-inc.com
from __future__ import annotations

import asyncio
import json
from uuid import uuid4

import requests
import websockets
from consys4py.constants import APIResourceTypes
from consys4py.datamodels.observations import ObservationOMJSONInline
from consys4py.datamodels.swe_components import DataRecordSchema

from .core_datamodels import DatastreamResource, SystemResource, TimePeriod
from .timemanagement import TemporalModes


# from swecommondm.component_implementations import DataRecord


class DataStream:
    """
    DataSource: represents the active connection of a datastream object.
    This class may later be used to connect to a control channel as well. It will almost certainly be used
    for Control Stream status monitoring.

    Attributes:
        name: Human readable name of the DataSource
        _datastream: DatastreamResource object
        _parent_system: System object that the DataSource is associated with.
    """
    name: str = None
    _id: str = None
    _datastream: DatastreamResource = None
    _parent_system: SystemResource = None
    _playback_mode: TemporalModes = None
    _url: str = None
    _auth: str = None
    _playback_websocket: websockets.WebSocketClientProtocol = None
    _extra_headers: dict = None
    _result_schema: DataRecordSchema = None

    def __init__(self, name: str, datastream: DatastreamResource,
                 parent_system: SystemResource):
        """
        :param name: Human-readable name of the DataSource
        :param datastream: DatastreamResource object
        :param parent_system: System object that the DataSource is associated with.
        """
        self._status = None
        self._id = f'datasource-{uuid4()}'
        self.name = name
        self._datastream = datastream
        self._playback_websocket = None
        self._parent_system = parent_system
        self._playback_mode = None
        self._url = None
        self._auth = None
        self._extra_headers = None
        if self._parent_system.get_parent_node().is_secure:
            self._auth = self._parent_system.get_parent_node().get_decoded_auth()
            self._extra_headers = {'Authorization': f'Basic {self._auth}'}
            # get result schema

            # t_url = f'http://{self._parent_system.get_parent_node().get_address()}:{self._parent_system.get_parent_node().get_port()}'
            #
            # res = consys4py.part_2.datastreams.retrieve_datastream_schema(t_url,
            #                                                               datastream_id=self._datastream.ds_id,
            #                                                               api_root=self._parent_system.get_parent_node()._api_helper.api_root,
            #                                                               headers=self._extra_headers)
            # print(res.json())

    def get_id(self) -> str:
        """
        Get the ID of the DataSource

        :return: str UID of the DataSource
        """
        return self._id

    def get_name(self):
        """
        Get the name of the DataSource

        :return: str name of the DataSource
        """
        return self.name

    def create_process(self):
        """
        **Unimplemented**

        Create a process for the DataSource

        :return:
        """
        pass

    def terminate_process(self):
        """
        **Unimplemented**
        """
        pass

    # Might not be necessary
    def subscribe(self):
        """
        **Unimplemented**

        :return:
        """
        pass

    def set_playback_mode(self, mode: TemporalModes):
        """
        Sets the playback mode of the DataSource and regenerates the URL accordingly

        :param mode: TemporalModes

        :return:
        """
        self._playback_mode = mode
        self.generate_retrieval_url()

    def initialize(self):
        """
        Initializes the DataSource object, resetting the status and closing any open connections if necessary.

        :return:
        """
        if self._playback_websocket.is_open():
            self._playback_websocket.close()
        self._playback_websocket = None
        self._status = "initialized"

    async def connect(self) -> websockets.WebSocketClientProtocol or None:
        """
        Attempts to connect to the DataSource's websocket, or HTTP endpoint if in BATCH mode. This is currently for retrieval
        :return: The websocket connection if in REAL_TIME or ARCHIVE mode, ``None`` if in BATCH mode.
        """
        if self._playback_mode == TemporalModes.REAL_TIME:
            self._playback_websocket = await websockets.connect(self._url,
                                                                extra_headers=self._extra_headers)
            self._status = "connected"
            return self._playback_websocket
        elif self._playback_mode == TemporalModes.ARCHIVE:
            self._playback_websocket = await websockets.connect(self._url,
                                                                extra_headers=self._extra_headers)
            self._status = "connected"
            return self._playback_websocket
        elif self._playback_mode == TemporalModes.BATCH:
            self._playback_websocket = await websockets.connect(self._url,
                                                                extra_headers=self._extra_headers)
            self._status = "connected"
            return None

    def disconnect(self):
        """
        Closes the websocket connection, *should* also stop any future http requests if in BATCH mode. This feature
        is *WIP*.

        :return:
        """
        self._playback_websocket.close()

    def reset(self):
        """
        Resets the DataSource object, closing any open connections and resetting the status. Currently has the same
        effect as ``initialize()``.

        :return:
        """
        if self._playback_websocket.is_open():
            self._playback_websocket.close()
        self._playback_websocket = None
        self._status = "initialized"

    def get_status(self):
        """
        Get the status code of the DataSource

        :return:
        """
        return self._status

    def get_parent_system(self) -> SystemResource:
        """
        Retrieve the DataSource's parent System

        :return: The parent System object of the DataSource
        """
        return self._parent_system

    def get_ws_client(self):
        """
        Get the websocket client object

        :return:
        """
        return self._playback_websocket

    def is_within_timeperiod(self, timeperiod: TimePeriod) -> bool:
        """
        Checks if the DataSource's Datastream is within the provided TimePeriod

        :param timeperiod: TimePeriod object
        :return: ``True`` if the Datastream is within the TimePeriod, ``False`` otherwise
        """
        return timeperiod.does_timeperiod_overlap(self._datastream.valid_time)

    def generate_retrieval_url(self):
        """
        Generates the URL for the DataSource based on the playback mode. This url is used for accessing the datastream
        on the OSH server.

        :return:
        """
        # TODO: need to specify secure vs insecure protocols
        if self._playback_mode == TemporalModes.REAL_TIME:
            self._url = (
                f'ws://{self._parent_system.get_parent_node().get_address()}:'
                f'{self._parent_system.get_parent_node().get_port()}'
                f'/sensorhub/api/datastreams/{self._datastream.ds_id}'
                f'/observations?f=application%2Fjson')
        elif self._playback_mode == TemporalModes.ARCHIVE:
            self._url = (
                f'ws://{self._parent_system.get_parent_node().get_address()}:'
                f'{self._parent_system.get_parent_node().get_port()}'
                f'/sensorhub/api/datastreams/{self._datastream.ds_id}'
                f'/observations?f=application%2Fjson&resultTime={self._datastream.valid_time.start}/'
                f'{self._datastream.valid_time.end}')
        elif self._playback_mode == TemporalModes.BATCH:
            # TODO: need to allow for batch counts selection through DS Handler or TimeManager
            self._url = (
                f'wss://{self._parent_system.get_parent_node().get_address()}:'
                f'{self._parent_system.get_parent_node().get_port()}'
                f'/sensorhub/api/datastreams/{self._datastream.ds_id}'
                f'/observations?f=application%2Fjson&resultTime={self._datastream.valid_time.start}/'
                f'{self._datastream.valid_time.end}')
        else:
            raise ValueError(
                "Playback mode not set. Cannot generate URL for DataSource.")

    def generate_insertion_url(self) -> str:
        """
        Generates the URL for the DataSource. This url is used for posting data to the
        OSH server.

        :return:
        """
        url_result = (
            f'http://{self._parent_system.get_parent_node().get_address()}:'
            f'{self._parent_system.get_parent_node().get_port()}'
            f'/sensorhub/api/datastreams/{self._datastream.ds_id}'
            f'/observations'
        )
        return url_result

    def insert_observation(self, observation: ObservationOMJSONInline):
        """
        Posts an observation to the server
        :param observation: ObservationOMJSONInline object
        :return:
        """
        api_helper = self._parent_system.get_parent_node().get_api_helper()
        api_helper.post_resource(APIResourceTypes.OBSERVATION, parent_res_id=self._datastream.ds_id,
                                 data=observation.model_dump(), req_headers={'Content-Type': 'application/om+json'})


class DataStreamHandler:
    """
    Manages a collection of DataSource objects, allowing for easy access and control of multiple datastreams. As well
    as providing them access to a message handler for processing incoming data.
    """
    datasource_map: dict[str, DataStream]
    _message_list: MessageHandler
    _playback_mode: TemporalModes

    def __init__(self, playback_mode: TemporalModes = TemporalModes.REAL_TIME):
        self.datasource_map = {}
        self._message_list = MessageHandler()
        self._playback_mode = playback_mode

    def set_playback_mode(self, mode: TemporalModes):
        """
        Sets the playback mode for the DataSourceHandler and all of its DataSources

        :param mode: TemporalModes

        :return:
        """
        self._playback_mode = mode

    def add_datasource(self, datasource: DataStream):
        """
        Adds a DataSource object to the DataSourceHandler

        :param datasource: DataSource

        :return:
        """
        datasource.set_playback_mode(self._playback_mode)
        self.datasource_map[datasource.get_id()] = datasource

    def remove_datasource(self, datasource_id: str) -> DataStream:
        """
        Removes a DataSource object from the DataSourceHandler

        :param datasource_id: str uid of the DataSource

        :return: the removed DataSource object
        """
        return self.datasource_map.pop(datasource_id)

    def initialize_ds(self, datasource_id: str):
        """
        Initializes a DataSource object by calling its initialize method

        :param datasource_id:

        :return:
        """
        ds = self.datasource_map.get(datasource_id)
        ds.initialize()

    def initialize_all(self):
        """
        Initializes all DataSource objects in the DataSourceHandler

        :return:
        """
        [ds.initialize() for ds in self.datasource_map.values()]

    def set_ds_mode(self):
        """
        Sets the playback mode for all DataSource objects in the DataSourceHandler, uses the playback mode of the
        DataSourceHandler
        :return:
        """
        (ds.set_playback_mode(self._playback_mode) for ds in self.datasource_map.values())

    async def connect_ds(self, datasource_id: str):
        """
        Connects a DataSource object by calling its connect method

        :param datasource_id:

        :return:
        """
        ds = self.datasource_map.get(datasource_id)
        await ds.connect()

    async def connect_all(self, timeperiod: TimePeriod):
        """
        Connects all datasources, optionally within a provided TimePeriod
        :param timeperiod: TimePeriod object
        :return:
        """
        # search for datasources that fall within the timeperiod
        if timeperiod is not None:
            ds_matches = [ds for ds in self.datasource_map.values() if
                          ds.is_within_timeperiod(timeperiod)]
        else:
            ds_matches = self.datasource_map.values()

        if self._playback_mode == TemporalModes.REAL_TIME:
            [(ds, await ds.connect()) for ds in ds_matches]
            for ds in ds_matches:
                asyncio.create_task(self._handle_datastream_client(ds))
        elif self._playback_mode == TemporalModes.ARCHIVE:
            pass
        elif self._playback_mode == TemporalModes.BATCH:
            for ds in ds_matches:
                asyncio.create_task(self.handle_http_batching(ds))

    def disconnect_ds(self, datasource_id: str):
        """
        Disconnects a DataSource object by calling its disconnect method
        :param datasource_id:
        :return:
        """
        ds = self.datasource_map.get(datasource_id)
        ds.disconnect()

    def disconnect_all(self):
        """
        Disconnects all DataSource objects in the DataSourceHandler
        :return:
        """
        [ds.disconnect() for ds in self.datasource_map.values()]

    async def _handle_datastream_client(self, datasource: DataStream):
        """
        Handles the websocket client for a DataSource object, passes Observations to the MessageHandler in the
        form of MessageWrapper objects

        :param datasource:

        :return:
        """
        try:
            async for msg in datasource.get_ws_client():
                msg_dict = json.loads(msg.decode('utf-8'))
                obs = ObservationOMJSONInline.model_validate(msg_dict)
                msg_wrapper = MessageWrapper(datasource=datasource,
                                             message=obs)
                self._message_list.add_message(msg_wrapper)

        except Exception as e:
            print(f"An error occurred while reading from websocket: {e}")

    async def handle_http_batching(self, datasource: DataStream,
                                   offset: int = None,
                                   query_params: dict = None,
                                   next_link: str = None) -> dict:
        """
        Handles the batching of HTTP requests for a DataSource object, passes Observations to the MessageHandler

        :param datasource:
        :param offset:
        :param query_params:
        :param next_link:

        :return: dict of the response from the server
        """
        # access api_helper
        api_helper = datasource.get_parent_system().get_parent_node().get_api_helper()
        # needs to create a new call to make a request to the server if there is a link to a next page
        resp = None
        if next_link is None:
            resp = api_helper.retrieve_resource(APIResourceTypes.OBSERVATION,
                                                parent_res_id=datasource._datastream.ds_id,
                                                req_headers={
                                                    'Content-Type': 'application/json'})
        elif next_link is not None:
            resp = requests.get(next_link, auth=(
                datasource._parent_system.get_parent_node()._api_helper.username,
                datasource._parent_system.get_parent_node()._api_helper.password))
        results = resp.json()
        if 'links' in results:
            for link in results['links']:
                if link['rel'] == 'next':
                    # new_offset = link['href'].split('=')[-1]
                    asyncio.create_task(self.handle_http_batching(datasource, next_link=link['href']))

        # print(results)
        for obs in results['items']:
            obs_obj = ObservationOMJSONInline.model_validate(obs)
            msg_wrapper = MessageWrapper(datasource=datasource,
                                         message=obs_obj)
            self._message_list.add_message(msg_wrapper)
        return resp.json()

    def get_message_handler(self) -> MessageHandler:
        """
        Get the MessageHandler object from the DataSourceHandler

        :return: MessageHandler object
        """
        return self._message_list

    def get_messages(self) -> list[MessageWrapper]:
        """
        Get the list of MessageWrapper objects from the MessageHandler

        :return: List of MessageWrapper objects
        """
        return self._message_list.get_messages()

    def post_observation(self, datasource: DataStream, observation: ObservationOMJSONInline):
        """
        Posts an observation to the server

        :param datasource: DataSource object
        :param observation: ObservationOMJSONInline object

        :return:
        """
        api_helper = datasource.get_parent_system().get_parent_node().get_api_helper()
        api_helper.post_resource(APIResourceTypes.OBSERVATION, parent_res_id=datasource._datastream.ds_id,
                                 data=observation.model_dump(), req_headers={'Content-Type': 'application/json'})


class MessageHandler:
    """
    Manages a list of MessageWrapper objects, allowing for easy access and control of multiple messages. Works in
    conjunction with the TimeManager to sort messages by their resultTime.
    """
    _message_list: list[MessageWrapper]

    def __init__(self):
        self._message_list = []

    def add_message(self, message: MessageWrapper):
        """
        Adds a MessageWrapper object to the MessageHandler

        :param message:

        :return:
        """
        self._message_list.append(message)
        # print(self._message_list)

    def get_messages(self) -> list[MessageWrapper]:
        """
        Get the list of MessageWrapper objects

        :return: List of MessageWrapper objects
        """
        return self._message_list

    def clear_messages(self):
        """
        Empties the list of MessageWrapper objects

        :return:
        """
        self._message_list.clear()

    def sort_messages(self) -> list[MessageWrapper]:
        """
        Sorts the list of MessageWrapper objects by their resultTime

        :return: the sorted List of MessageWrapper objects
        """
        # copy the list
        sorted_list = self._message_list.copy()
        sorted_list.sort(key=lambda x: x.resultTime)
        return sorted_list


class MessageWrapper:
    """
    Combines a DataSource and a Message into a single object for easier access
    """

    def __init__(self, datasource: DataStream,
                 message: ObservationOMJSONInline):
        self._message = message
        self._datasource = datasource

    def get_message(self) -> ObservationOMJSONInline:
        """
        Get the observation data from the MessageWrapper

        :return: ObservationOMJSONInline that is easily serializable
        """
        return self._message

    def get_message_as_dict(self) -> dict:
        """
        Get the observation data from the MessageWrapper as a dictionary

        :return: dict of the observation result data
        """
        return self._message.model_dump()

    def __repr__(self):
        return f"{self._datasource}, {self._message}"



#   ==============================================================================
#   Copyright (c) 2024 Botts Innovative Research, Inc.
#   Date:  2024/6/26
#   Author:  Ian Patterson
#   Contact Email:  ian@botts-inc.com
#   ==============================================================================
from __future__ import annotations

from typing import List

from consys4py.datamodels.swe_components import Geometry
from consys4py.datamodels.datastreams import DatastreamSchema
from consys4py.datamodels.api_utils import Link
from pydantic import BaseModel, ConfigDict, Field, SerializeAsAny
from shapely import Point

from oshconnect.timemanagement import DateTimeSchema, TimePeriod


class BoundingBox(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    lower_left_corner: Point = Field(..., description="The lower left corner of the bounding box.")
    upper_right_corner: Point = Field(..., description="The upper right corner of the bounding box.")
    min_value: float = Field(None, description="The minimum value of the bounding box.")
    max_value: float = Field(None, description="The maximum value of the bounding box.")

    # @model_validator(mode='before')
    # def validate_minmax(self) -> Self:
    #     if self.min_value > self.max_value:
    #         raise ValueError("min_value must be less than max_value")
    #     return self


class SecurityConstraints:
    constraints: list


class LegalConstraints:
    constraints: list


class Characteristics:
    characteristics: list


class Capabilities:
    capabilities: list


class Contact:
    contact: list


class Documentation:
    documentation: list


class HistoryEvent:
    history_event: list


class ConfigurationSettings:
    settings: list


class FeatureOfInterest:
    feature: list


class Input:
    input: list


class Output:
    output: list


class Parameter:
    parameter: list


class Mode:
    mode: list


class ProcessMethod:
    method: list


class SystemResource(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    feature_type: str = Field(None, serialization_alias="type")
    system_id: str = Field(None, serialization_alias="id")
    properties: dict = Field(None)
    geometry: Geometry | None = Field(None)
    bbox: BoundingBox = Field(None)
    links: List[Link] = Field(None)
    description: str = Field(None)
    uid: str = Field(None, serialization_alias="uniqueId")
    label: str = Field(None)
    lang: str = Field(None)
    keywords: List[str] = Field(None)
    identifiers: List[str] = Field(None)
    classifiers: List[str] = Field(None)
    valid_time: DateTimeSchema = Field(None, serialization_alias="validTime")
    security_constraints: List[SecurityConstraints] = Field(None, serialization_alias="securityConstraints")
    legal_constraints: List[LegalConstraints] = Field(None, serialization_alias="legalConstraints")
    characteristics: List[Characteristics] = Field(None)
    capabilities: List[Capabilities] = Field(None)
    contacts: List[Contact] = Field(None)
    documentation: List[Documentation] = Field(None)
    history: List[HistoryEvent] = Field(None)
    definition: str = Field(None)
    type_of: str = Field(None, serialization_alias="typeOf")
    configuration: ConfigurationSettings = Field(None)
    features_of_interest: List[FeatureOfInterest] = Field(None, serialization_alias="featuresOfInterest")
    inputs: List[Input] = Field(None)
    outputs: List[Output] = Field(None)
    parameters: List[Parameter] = Field(None)
    modes: List[Mode] = Field(None)
    method: ProcessMethod = Field(None)


class DatastreamResource(BaseModel):
    """
    The DatastreamResource class is a Pydantic model that represents a datastream resource in the OGC SensorThings API.
    It contains all the necessary and optional properties listed in the OGC Connected Systems API documentation. Note
    that, depending on the format of the  request, the fields needed may differ. There may be derived models in a later
    release that will have different sets of required fields to ease the validation process for users.
    """
    # model_config = ConfigDict(populate_by_name=True)

    ds_id: str = Field(..., serialization_alias="id")
    name: str = Field(...)
    description: str = Field(None)
    valid_time: TimePeriod = Field(..., serialization_alias="validTime")
    output_name: str = Field(None, serialization_alias="outputName")
    procedure_link: Link = Field(None, serialization_alias="procedureLink@link")
    deployment_link: Link = Field(None, serialization_alias="deploymentLink@link")
    ultimate_feature_of_interest_link: Link = Field(None, serialization_alias="ultimateFeatureOfInterest@link")
    sampling_feature_link: Link = Field(None, serialization_alias="samplingFeature@link")
    parameters: dict = Field(None)
    phenomenon_time: TimePeriod = Field(None, serialization_alias="phenomenonTimeInterval")
    result_time: TimePeriod = Field(None, serialization_alias="resultTimeInterval")
    ds_type: str = Field(None, serialization_alias="type")
    result_type: str = Field(None, serialization_alias="resultType")
    links: List[Link] = Field(None)
    record_schema: SerializeAsAny[DatastreamSchema] = Field(None, serialization_alias="schema")


class ObservationResource(BaseModel):
    sampling_feature_id: str = Field(None, serialization_alias="samplingFeature@Id")
    procedure_link: Link = Field(None, serialization_alias="procedure@link")
    phenomenon_time: DateTimeSchema = Field(None, serialization_alias="phenomenonTime")
    result_time: DateTimeSchema = Field(..., serialization_alias="resultTime")
    parameters: dict = Field(None)
    result: dict = Field(...)
    result_link: Link = Field(None, serialization_alias="result@link")



#   ==============================================================================
#   Copyright (c) 2024 Botts Innovative Research, Inc.
#   Date:  2024/6/26
#   Author:  Ian Patterson
#   Contact Email:  ian@botts-inc.com
#   ==============================================================================

from __future__ import annotations

import re
import time
from datetime import datetime, timezone
from enum import Enum
from typing import Any, Self

from pydantic import BaseModel, ConfigDict, Field, field_validator, model_serializer, model_validator


class TemporalModes(Enum):
    REAL_TIME = "realtime"
    ARCHIVE = "archive"
    BATCH = "batch"
    RT_SYNC = "realtimesync"
    ARCHIVE_SYNC = "archivesync",


class State(Enum):
    UNINITIALIZED = 0
    INITIALIZED = 1
    STOPPED = 2
    BUFFERING = 3
    PLAYING = 4
    FAST_FORWARDING = 5
    REWINDING = 6


class TimeUtils:
    iso_format = '%Y-%m-%dT%H:%M:%S.%fZ'

    @staticmethod
    def to_epoch_time(a_time: datetime | str) -> float:
        """
        Convert a datetime or string to epoch time
        :param a_time:
        :return:
        """
        if isinstance(a_time, str):
            return time.mktime(
                datetime.strptime(a_time, "%Y-%m-%d %H:%M:%S.%fZ").timetuple())
        elif isinstance(a_time, datetime):
            return time.mktime(a_time.timetuple())

    @staticmethod
    def to_utc_time(a_time: float | str) -> datetime:
        """
        Convert epoch time or string to UTC time object
        :param a_time:
        :return:
        """
        if isinstance(a_time, str):
            if re.match(r'(\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.)(\d+)(Z)', a_time):
                return datetime.strptime(a_time, "%Y-%m-%dT%H:%M:%S.%fZ").replace(tzinfo=timezone.utc)
            else:
                return datetime.strptime(a_time, "%Y-%m-%dT%H:%M:%SZ").replace(tzinfo=timezone.utc)
        elif isinstance(a_time, float):
            return datetime.fromtimestamp(a_time, tz=timezone.utc)

    @staticmethod
    def current_epoch_time():
        """
        Get the current time in epoch format
        :return:
        """
        return time.time()

    @staticmethod
    def current_utc_time() -> datetime:
        """
        Get the current time in UTC timezone
        :return:
        """
        return datetime.now(timezone.utc)

    @staticmethod
    def time_to_iso(a_time: datetime | float) -> str:
        """
        Convert a datetime object to iso format
        :param a_time: datetime object in UTC timezone or epoch time (float)
        :return:
        """
        if isinstance(a_time, float):
            return datetime.fromtimestamp(a_time).strftime(TimeUtils.iso_format)
        elif isinstance(a_time, datetime):
            return a_time.strftime(TimeUtils.iso_format)

    @staticmethod
    def compare_time_instants_or_indeterminate(time1: TimeInstant | str, time2: TimeInstant | str) -> int:
        """
        Compare two time instants or indeterminate times. This coerces the indeterminate time 'now' to the current time.
        This may cause unexpected behavior if the times are very close together.
        :param time1: TimeInstant or IndeterminateTime
        :param time2: TimeInstant or IndeterminateTime
        :return: 0 if equal, -1 if time1 < time2, 1 if time1 > time2
        """
        if isinstance(time1, TimeInstant) and isinstance(time2, TimeInstant):
            if time1 < time2:
                return -1
            elif time1 > time2:
                return 1
            else:
                return 0
        elif isinstance(time1, TimeInstant) and isinstance(time2, str):
            if time2 == "now":
                now_as_time = TimeInstant.now_as_time_instant()
                return TimeUtils.compare_time_instants_or_indeterminate(time1, now_as_time)
            else:
                t2_as_time = TimeInstant.from_string(time2)
                return TimeUtils.compare_time_instants_or_indeterminate(time1, t2_as_time)
        elif isinstance(time1, str) and isinstance(time2, TimeInstant):
            if time1 == "now":
                now_as_ti = TimeInstant.now_as_time_instant()
                return TimeUtils.compare_time_instants_or_indeterminate(now_as_ti, time2)
            else:
                t1_as_ti = TimeInstant.from_string(time1)
                return TimeUtils.compare_time_instants_or_indeterminate(t1_as_ti, time2)
        elif isinstance(time1, str) and isinstance(time2, str):
            if time1 == "now" and time2 == "now":
                raise ValueError("Both times cannot be 'now'")
            elif time1 == "now":
                t1_as_ti = TimeInstant.now_as_time_instant()
                t2_as_ti = TimeInstant.from_string(time2)
            elif time2 == "now":
                t2_as_ti = TimeInstant.now_as_time_instant()
                t1_as_ti = TimeInstant.from_string(time1)
            else:
                t1_as_ti = TimeInstant.from_string(time1)
                t2_as_ti = TimeInstant.from_string(time2)
            return TimeUtils.compare_time_instants_or_indeterminate(t1_as_ti, t2_as_ti)


class TimeInstant:
    _epoch_time: float | None

    def __init__(self, epoch_time: float = None, utc_time: datetime = None):
        if epoch_time is not None:
            self._epoch_time = epoch_time
        elif utc_time is not None:
            self._epoch_time = TimeUtils.to_epoch_time(utc_time)

    @property
    def epoch_time(self):
        return self._epoch_time

    @epoch_time.setter
    def epoch_time(self, epoch_time: float):
        if hasattr(self, "_epoch_time"):
            raise AttributeError("Epoch time should not be changed once set")

    def has_epoch_time(self):
        return self._epoch_time is not None

    def get_utc_time(self):
        return TimeUtils.to_utc_time(self._epoch_time)

    def get_iso_time(self):
        return TimeUtils.time_to_iso(self._epoch_time)

    @staticmethod
    def from_string(utc_time: str):
        # TODO: handle timezones
        if re.match(r'(\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.)(\d+)(Z)', utc_time):
            dt = datetime.strptime(utc_time, "%Y-%m-%dT%H:%M:%S.%fZ")
        else:
            dt = datetime.strptime(utc_time, "%Y-%m-%dT%H:%M:%SZ")
        return TimeInstant(utc_time=dt)

    @staticmethod
    def now_as_time_instant():
        return TimeInstant(epoch_time=TimeUtils.current_epoch_time())

    def __lt__(self, other: TimeInstant) -> bool:
        return self.epoch_time < other.epoch_time

    def __gt__(self, other: TimeInstant) -> bool:
        return self.epoch_time > other.epoch_time

    def __eq__(self, other: TimeInstant) -> bool:
        return self.epoch_time == other.epoch_time

    def __le__(self, other: TimeInstant) -> bool:
        return self.epoch_time <= other.epoch_time

    def __ge__(self, other: TimeInstant) -> bool:
        return self.epoch_time >= other.epoch_time

    def __ne__(self, other: TimeInstant) -> bool:
        return self.epoch_time != other.epoch_time

    def __repr__(self):
        return f'{self.get_iso_time()}'


class DateTimeSchema(BaseModel):
    is_instant: bool = Field(True, description="Whether the date time is an instant or a period.")
    iso_date: str = Field(None, description="The ISO formatted date time.")
    time_period: tuple = Field(None, description="The time period of the date time.")

    @model_validator(mode='before')
    def valid_datetime_type(self) -> Self:
        if self.is_instant:
            if self.iso_date is None:
                raise ValueError("Instant date time must have a valid ISO8601 date.")
        return self

    @field_validator('iso_date')
    @classmethod
    def check_iso_date(cls, v) -> str:
        if not v:
            raise ValueError("Instant date time must have a valid ISO8601 date.")
        return v


class IndeterminateTime(Enum):
    NOW = "now"
    LATEST = "latest"
    FIRST = "first"


class TimePeriod(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    start: TimeInstant | str = Field(...)
    end: TimeInstant | str = Field(...)

    @model_validator(mode='before')
    @classmethod
    def valid_time_period(cls, data) -> Any:
        data_dict = {}

        if isinstance(data, list):
            data_dict['start'] = cls.check_mbr_type(data[0])
            data_dict['end'] = cls.check_mbr_type(data[1])
        elif isinstance(data, dict):
            data_dict['start'] = cls.check_mbr_type(data['start'])
            data_dict['end'] = cls.check_mbr_type(data['end'])

        if not cls.compare_start_lt_end(data_dict['start'], data_dict['end']):
            raise ValueError("Start time must be less than end time")

        return data_dict

    @model_serializer
    def ser_model(self):
        return [str(self.start), str(self.end)]

    @staticmethod
    def check_mbr_type(value):
        if isinstance(value, str):
            if value == "now":
                return value
            else:
                tp = TimeInstant.from_string(value)
                return tp
        elif isinstance(value, TimeInstant):
            return value

    @classmethod
    def compare_start_lt_end(cls, start: TimeInstant | str, end: TimeInstant | str) -> bool:
        if isinstance(start, TimeInstant) and isinstance(end, TimeInstant):
            return start < end
        elif isinstance(start, str) and isinstance(end, str):
            if start == "now" and end == "now":
                raise ValueError("Start and end cannot both be 'now'")
            else:
                raise ValueError("Check the time strings for validity. This should not occur.")
        elif isinstance(start, str) and start == "now" and isinstance(end, TimeInstant):
            return TimeInstant.now_as_time_instant() < end
        elif isinstance(start, TimeInstant) and isinstance(end, str) and end == "now":
            return start < TimeInstant.now_as_time_instant()

    def __repr__(self):
        return f'{[self.start, self.end]}'

    def does_timeperiod_overlap(self, checked_timeperiod: TimePeriod) -> bool:
        """
        Checks if the provided TimePeriod overlaps with the TimePeriod instance.

        **Note**: This method does not check for some edge cases, but the TimePeriods *should never* be valid
        in those situations.

        :param checked_timeperiod:
        :return: True if the TimePeriods overlap, False otherwise
        """
        # check that start of checked is not after end of this instance
        start_check_lt_end_inst = TimeUtils.compare_time_instants_or_indeterminate(checked_timeperiod.start, self.end)
        # check that end of checked is not before start of this instance
        end_check_gt_start_inst = TimeUtils.compare_time_instants_or_indeterminate(self.end, checked_timeperiod.start)
        if start_check_lt_end_inst == -1 and end_check_gt_start_inst == 1:
            return True


class Utilities:
    pass
    # @staticmethod
    # def parse_systems_result(result) -> System:


class TimeManagement:
    time_range: TimePeriod
    time_controller: TimeController

    def __init__(self, time_range: TimePeriod):
        self.time_range = time_range
        self.time_controller = TimeController()

    def get_time_range(self):
        return self.time_range


class TemporalMode:
    pass


class TimeController:
    _instance = None
    _temporal_mode: TemporalMode
    _status: str
    _playback_speed: int
    _timeline_begin: TimeInstant
    _timeline_end: TimeInstant
    _current_time: TimeInstant

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(TimeController, cls).__new__(cls)
        return cls._instance

    def set_temporal_mode(self, mode: TemporalMode):
        self._temporal_mode = mode

    def get_temporal_mode(self):
        return self._temporal_mode

    def start(self):
        pass

    def pause(self):
        pass

    def stop(self):
        pass

    def fast_forward(self, speed: int):
        self._playback_speed = speed

    def rewind(self, speed: int):
        self._playback_speed = speed

    def skip(self, a_time: TimeInstant):
        self._current_time = a_time

    def get_status(self):
        return self._status

    def set_timeline_start(self, a_time: TimeInstant):
        self._timeline_begin = a_time

    def set_timeline_end(self, a_time: TimeInstant):
        self._timeline_end = a_time

    def set_current_time(self, a_time: TimeInstant):
        if a_time < self._timeline_begin:
            self._current_time = self._timeline_begin
        elif a_time > self._timeline_end:
            self._timeline_end = a_time
        self._current_time = a_time

    def get_timeline_start(self):
        return self._timeline_begin

    def get_timeline_end(self):
        return self._timeline_end

    def get_current_time(self):
        return self._current_time

    def play_from_start(self):
        pass

    def skip_to_end(self):
        pass

    def add_listener(self, datastream, event_listener) -> str:
        pass

    def remove_listener(self, stream_id):
        pass

    def clear_streams(self):
        pass

    def reset(self):
        self.clear_streams()
        self._temporal_mode = None
        self._status = None
        self._playback_speed = None
        self._timeline_begin = None
        self._timeline_end = None
        self._current_time = None

    def set_buffer_time(self, time: int):
        pass

    def get_buffer_time(self):
        pass

    def _compute_time_range(self):
        pass


class Synchronizer:
    _buffer: any

    def synchronize(self, systems: list):
        pass



#   ==============================================================================
#   Copyright (c) 2024 Botts Innovative Research, Inc.
#   Date:  2024/6/26
#   Author:  Ian Patterson
#   Contact Email:  ian@botts-inc.com
#   ==============================================================================
from __future__ import annotations

import base64
import uuid
from dataclasses import dataclass, field

from consys4py.constants import APIResourceTypes
from consys4py.core.default_api_helpers import APIHelper
from consys4py.datamodels.datastreams import SWEDatastreamSchema
from consys4py.datamodels.encoding import JSONEncoding
from consys4py.datamodels.swe_components import DataRecordSchema

from .core_datamodels import DatastreamResource, ObservationResource, SystemResource
from .timemanagement import TimeInstant, TimePeriod, TimeUtils


@dataclass(kw_only=True)
class Endpoints:
    root: str = "sensorhub"
    sos: str = f"{root}/sos"
    connected_systems: str = f"{root}/api"


class Utilities:

    @staticmethod
    def convert_auth_to_base64(username: str, password: str) -> str:
        return base64.b64encode(f"{username}:{password}".encode()).decode()


@dataclass(kw_only=True)
class Node:
    _id: str
    protocol: str
    address: str
    port: int
    endpoints: Endpoints
    is_secure: bool
    _basic_auth: bytes = None
    _api_helper: APIHelper
    # _system_ids: list[uuid] = field(default_factory=list)
    _systems: list[System] = field(default_factory=list)

    def __init__(self, protocol: str, address: str, port: int,
                 username: str = None, password: str = None,
                 **kwargs: dict):
        self._id = f'node-{uuid.uuid4()}'
        self.protocol = protocol
        self.address = address
        self.port = port
        self.is_secure = username is not None and password is not None
        if self.is_secure:
            self.add_basicauth(username, password)
        self.endpoints = Endpoints()
        self._api_helper = APIHelper(
            server_url=f'{self.protocol}://{self.address}:{self.port}',
            api_root=self.endpoints.connected_systems, username=username,
            password=password)
        if self.is_secure:
            self._api_helper.user_auth = True
        self._systems = []

    def get_id(self):
        return self._id

    def get_address(self):
        return self.address

    def get_port(self):
        return self.port

    def get_api_endpoint(self):
        return f"http{'s' if self.is_secure else ''}://{self.address}:{self.port}/{self.endpoints.connected_systems}"

    def add_basicauth(self, username: str, password: str):
        if not self.is_secure:
            self.is_secure = True
        self._basic_auth = base64.b64encode(
            f"{username}:{password}".encode('utf-8'))

    def get_decoded_auth(self):
        return self._basic_auth.decode('utf-8')

    def discover_systems(self):
        result = self._api_helper.retrieve_resource(APIResourceTypes.SYSTEM,
                                                    req_headers={})
        if result.ok:
            new_systems = []
            system_objs = result.json()['items']
            print(system_objs)
            for system_json in system_objs:
                print(system_json)
                system = SystemResource.model_validate(system_json)
                sys_obj = System.from_system_resource(system)
                sys_obj.update_parent_node(self)
                self._systems.append(sys_obj)
                new_systems.append(sys_obj)
            return new_systems
        else:
            return None

    def add_new_system(self, system: System):
        system.update_parent_node(self)
        self._systems.append(system)

    def get_api_helper(self) -> APIHelper:
        return self._api_helper

    # System Management

    def add_system(self, system: System, target_node: Node, insert_resource: bool = False):
        """
        Add a system to the target node.
        :param system: System object
        :param target_node: Node object
        :param insert_resource: Whether to insert the system into the target node's server, default is False
        :return:
        """
        if insert_resource:
            system.insert_self()
        target_node.add_new_system(system)
        self._systems.append(system)
        return system

    def systems(self) -> list[System]:
        return self._systems


class System:
    uid: uuid.UUID
    resource_id: str
    name: str
    label: str
    # datastreams: list[Datastream]
    datastreams: list[DatastreamResource]
    control_channels: list[ControlChannel]
    description: str
    urn: str
    _parent_node: Node
    _sys_resource: SystemResource

    def __init__(self, name: str, label: str, urn: str, **kwargs):
        """
        :param name: The machine-accessible name of the system
        :param label: The human-readable label of the system
        :param urn: The URN of the system, typically formed as such: 'urn:general_identifier:specific_identifier:more_specific_identifier'
        :param kwargs:
            - 'description': A description of the system
        """
        self.uid = uuid.uuid4()
        self.name = name
        self.label = label
        self.datastreams = []
        self.control_channels = []
        self.urn = urn
        if kwargs.get('resource_id'):
            self.resource_id = kwargs['resource_id']
        if kwargs.get('description'):
            self.description = kwargs['description']

    def update_parent_node(self, node: Node):
        self._parent_node = node

    def get_parent_node(self) -> Node:
        return self._parent_node

    def discover_datastreams(self) -> list[DatastreamResource]:
        res = self._parent_node.get_api_helper().retrieve_resource(
            APIResourceTypes.DATASTREAM, req_headers={})
        datastream_json = res.json()['items']
        ds_resources = []
        for ds in datastream_json:
            datastream_objs = DatastreamResource.model_validate(ds)
            ds_resources.append(datastream_objs)

        return ds_resources

    @staticmethod
    def from_system_resource(system_resource: SystemResource):
        other_props = system_resource.model_dump()
        print(f'Props of SystemResource: {other_props}')

        # case 1: has properties a la geojson
        if 'properties' in other_props:
            new_system = System(name=other_props['properties']['name'],
                                label=other_props['properties']['name'],
                                urn=other_props['properties']['uid'],
                                resource_id=system_resource.system_id)
        else:
            new_system = System(name=system_resource.name,
                                label=system_resource.label, urn=system_resource.urn,
                                resource_id=system_resource.system_id)
        return new_system

    def to_system_resource(self) -> SystemResource:
        resource = SystemResource(uid=self.urn, label=self.name, feature_type='PhysicalSystem')

        if len(self.datastreams) > 0:
            resource.outputs = [ds.to_resource() for ds in self.datastreams]

        # if len(self.control_channels) > 0:
        #     resource.inputs = [cc.to_resource() for cc in self.control_channels]
        return resource

    def add_insert_datastream(self, datastream: DataRecordSchema):
        """
        Adds a datastream to the system while also inserting it into the system's parent node via HTTP POST.
        :param datastream: DataRecordSchema to be used to define the datastream
        :return:
        """
        print(f'Adding datastream: {datastream.model_dump_json(exclude_none=True, by_alias=True)}')
        # Make the request to add the datastream
        # if successful, add the datastream to the system
        datastream_schema = SWEDatastreamSchema(record_schema=datastream, obs_format='application/swe+json',
                                                encoding=JSONEncoding())
        datastream_resource = DatastreamResource(ds_id="default", name=datastream.label, output_name=datastream.label,
                                                 record_schema=datastream_schema,
                                                 valid_time=TimePeriod(start=TimeInstant.now_as_time_instant(),
                                                                       end=TimeInstant(utc_time=TimeUtils.to_utc_time(
                                                                           "2026-12-31T00:00:00Z"))))

        api = self._parent_node.get_api_helper()
        print(
            f'Attempting to create datastream: {datastream_resource.model_dump_json(by_alias=True, exclude_none=True)}')
        print(
            f'Attempting to create datastream: {datastream_resource.model_dump(by_alias=True, exclude_none=True)}')
        res = api.create_resource(APIResourceTypes.DATASTREAM,
                                  datastream_resource.model_dump_json(by_alias=True, exclude_none=True),
                                  req_headers={
                                      'Content-Type': 'application/json'
                                  }, parent_res_id=self.resource_id)

        if res.ok:
            datastream_id = res.headers['Location'].split('/')[-1]
            print(f'Resource Location: {datastream_id}')
            datastream_resource.ds_id = datastream_id
        else:
            raise Exception(f'Failed to create datastream: {datastream_resource.name}')

        self.datastreams.append(datastream_resource)
        return Datastream(datastream_id, self._parent_node, datastream_resource)

    def insert_self(self):
        res = self._parent_node.get_api_helper().create_resource(
            APIResourceTypes.SYSTEM, self.to_system_resource().model_dump_json(by_alias=True, exclude_none=True),
            req_headers={
                'Content-Type': 'application/sml+json'
            })

        if res.ok:
            location = res.headers['Location']
            sys_id = location.split('/')[-1]
            self.resource_id = sys_id
            print(f'Created system: {self.resource_id}')

    def retrieve_resource(self):
        if self.resource_id is None:
            return None
        res = self._parent_node.get_api_helper().retrieve_resource(res_type=APIResourceTypes.SYSTEM,
                                                                   res_id=self.resource_id)
        if res.ok:
            system_json = res.json()
            print(system_json)
            system_resource = SystemResource.model_validate(system_json)
            print(f'System Resource: {system_resource}')
            self._sys_resource = system_resource


class Datastream:
    should_poll: bool
    _id: str
    _datastream_resource: DatastreamResource
    _parent_node: Node

    def __init__(self, id: str = None, parent_node: Node = None, datastream_resource: DatastreamResource = None):
        self._id = id
        self._parent_node = parent_node
        self._datastream_resource = datastream_resource

    def get_id(self):
        return self._datastream_resource.ds_id

    def insert_observation(self, observation: Observation):
        pass

    def to_resource(self) -> DatastreamResource:
        # if self._datastream_resource is None:
        #     self._datastream_resource = DatastreamResource(
        #         ds_id=uuid.uuid4(), name=self.name,
        #         valid_time=self.validTimeRange)
        return self._datastream_resource

    def observation_template(self) -> Observation:
        pass

    def create_observation(self, obs_data: dict):
        obs = ObservationResource(result=obs_data, result_time=TimeInstant.now_as_time_instant())
        # Validate against the schema
        if self._datastream_resource.record_schema is not None:
            obs.validate_against_schema(self._datastream_resource.record_schema)
        return obs

    def insert_observation_dict(self, obs_data: dict):
        res = self._parent_node.get_api_helper().create_resource(APIResourceTypes.OBSERVATION, obs_data,
                                                                 parent_res_id=self._id,
                                                                 req_headers={'Content-Type': 'application/json'})
        if res.ok:
            obs_id = res.headers['Location'].split('/')[-1]
            print(f'Inserted observation: {obs_id}')
            return id
        else:
            raise Exception(f'Failed to insert observation: {res.text}')

    # def create_from_record_schema(record_schema: DataRecordSchema, parent_system: System):
    #     new_ds = Datastream(name=record_schema.label, record_schema=record_schema)
    #     new_ds._datastream_resource = DatastreamResource(ds_id=uuid.uuid4(), name=new_ds.name)
    #     parent_system.datastreams.append(new_ds)
    #     return new_ds


class ControlChannel:
    # _cc_resource: ControlStream

    def __init__(self):
        pass


class Observation:
    _observation_resource: ObservationResource

    def __init__(self, observation_res: ObservationResource):
        self._observation_resource = observation_res

    def to_resource(self) -> ObservationResource:
        return self._observation_resource


class Output:
    name: str
    field_map: dict



#  ==============================================================================
#  Copyright (c) 2024. Botts Innovative Research, Inc.
#  Date:  2024/5/15
#  Author:  Ian Patterson
#  Contact email:  ian@botts-inc.com
#   ==============================================================================
import logging
import shelve

from consys4py.core.default_api_helpers import APIHelper

from .core_datamodels import DatastreamResource, TimePeriod
from .datasource import DataStream, DataStreamHandler, MessageWrapper
from .datastore import DataStore
from .osh_connect_datamodels import Node, System
from .styling import Styling
from .timemanagement import TemporalModes, TimeManagement


class OSHConnect:
    _name: str = None
    datastore: DataStore = None
    styling: Styling = None
    timestream: TimeManagement = None
    _nodes: list[Node] = []
    _systems: list[System] = []
    _cs_api_builder: APIHelper = None
    _datasource_handler: DataStreamHandler = None
    _datafeeds: list[DataStream] = []
    _datataskers: list[DataStore] = []
    _datagroups: list = []
    _tasks: list = []
    _playback_mode: TemporalModes = TemporalModes.REAL_TIME

    def __init__(self, name: str, **kwargs):
        """
        :param name: name of the OSHConnect instance, in the event that
        :param kwargs:
            - 'playback_mode': TemporalModes
        """
        self._name = name
        if 'nodes' in kwargs:
            self._nodes = kwargs['nodes']
            self._playback_mode = kwargs['playback_mode']
            self._datasource_handler.set_playback_mode(self._playback_mode)
        self._datasource_handler = DataStreamHandler()
        if 'playback_mode' in kwargs:
            self._playback_mode = kwargs['playback_mode']
            self._datasource_handler.set_playback_mode(self._playback_mode)

        logging.info(f"OSHConnect instance {name} created")

    def get_name(self):
        """
        Get the name of the OSHConnect instance.
        :return:
        """
        return self._name

    def add_node(self, node: Node):
        """
        Add a node to the OSHConnect instance.
        :param node: Node object
        :return:
        """
        self._nodes.append(node)

    def remove_node(self, node_id: str):
        """
        Remove a node from the OSHConnect instance.
        :param node_id:
        :return:
        """
        # TODO: should disconnect datastreams and delete them and all systems at the same time.
        # list of nodes in our node list that do not have the id of the node we want to remove
        self._nodes = [node for node in self._nodes if
                       node.get_id() != node_id]

    def save_config(self, config: dict):
        logging.info(f"Saving configuration for {self._name}")
        with shelve.open(f"{self._name}_config") as db:
            db['app_config'] = self
            db.close()

    @classmethod
    def load_config(cls, file_name: str) -> 'OSHConnect':
        with shelve.open(file_name, 'r') as db:
            app = db['app_config']
            db.close()
            return app

    def share_config(self, config: dict):
        pass

    def update_config(self, config: dict):
        pass

    def delete_config(self, config: dict):
        pass

    def configure_nodes(self, nodes: list):
        pass

    def filter_nodes(self, nodes: list):
        pass

    def task_system(self, task: dict):
        pass

    def select_temporal_mode(self, mode: str):
        """
        Select the temporal mode for the system. Real-time, archive, batch, as well as synchronization settings.
        :param mode:
        :return:
        """
        pass

    async def playback_streams(self, stream_ids: list = None):
        """
        Begins playback of the datastreams that have been connected to the app. The method of playback is determined
        by the temporal mode that has been set.
        :param stream_ids:
        :return:
        """
        if stream_ids is None:
            await self._datasource_handler.connect_all(
                self.timestream.get_time_range())
        else:
            for stream_id in stream_ids:
                await self._datasource_handler.connect_ds(stream_id)

    def visualize_streams(self, streams: list):
        pass

    # Second Level Use Cases
    def get_visualization_recommendations(self, streams: list):
        pass

    def discover_datastreams(self):
        """
        Discover datastreams of the current systems of the OSHConnect instance and create objects for them that are
        stored in the DataSourceHandler.
        :return:
        """
        # NOTE: This will need to check to prevent dupes in the future
        for system in self._systems:
            res_datastreams = system.discover_datastreams()
            # create DataSource(s)
            new_datasource = [
                DataStream(name=ds.name, datastream=ds, parent_system=system)
                for ds in
                res_datastreams]
            self._datafeeds.extend(new_datasource)
            list(map(self._datasource_handler.add_datasource, new_datasource))

    def discover_systems(self, nodes: list[str] = None):
        """
        Discover systems from the nodes that have been added to the OSHConnect instance. They are associated with the
        nodes that they are discovered from so access to them flows through there.
        :param nodes:
        :return:
        """
        search_nodes = self._nodes
        if nodes is not None:
            search_nodes = [node for node in search_nodes if
                            node.get_id() in nodes]

        for node in search_nodes:
            res_systems = node.discover_systems()
            self._systems.extend(res_systems)

    def discover_controlstreams(self, streams: list):
        pass

    def authenticate_user(self, user: dict):
        pass

    def synchronize_streams(self, systems: list):
        pass

    def set_playback_mode(self, mode: TemporalModes):
        self._datasource_handler.set_playback_mode(mode)

    def set_timeperiod(self, start_time: str, end_time: str):
        """
        Sets the time range (TimePeriod) for the OSHConnect instance. This is used to bookend the playback of the
        datastreams.
        :param start_time: ISO8601 formatted string or one of (now or latest)
        :param end_time:  ISO8601 formatted string or one of (now or latest)
        :return:
        """
        tp = TimePeriod(start=start_time, end=end_time)
        self.timestream = TimeManagement(time_range=tp)

    def get_message_list(self) -> list[MessageWrapper]:
        """
        Get the list of messages that have been received by the OSHConnect instance.
        :return: list of MessageWrapper objects
        """
        return self._datasource_handler.get_messages()

    def insert_system(self, system: System, target_node: Node):
        """
        Create a system on the target node.
        :param system: System object
        :param target_node: Node object, must be within the OSHConnect instance
        :return: the created system
        """
        if target_node in self._nodes:
            self.add_system(system, target_node, insert_resource=True)
            return system

    def insert_datastream(self, datastream: DatastreamResource, system: str | System) -> str:
        """
        Insert a datastream into the OSHConnect instance.
        :param datastream: DataSource object
        :param system: System object or system id
        :return:
        """
        sys_obj: System
        if isinstance(system, str):
            sys_obj = self.find_system(system)
            if sys_obj is None:
                raise ValueError(f"System with id {system} not found")
        else:
            sys_obj = system

        sys_obj.add_insert_datastream(datastream)

        self._datafeeds.append(datastream)

    def find_system(self, system_id: str) -> System | None:
        """
        Find a system in the OSHConnect instance.
        :param system_id:
        :return: the found system or None if not found
        """
        for system in self._systems:
            if system.uid == system_id:
                return system
        return None

    # System Management
    def add_system(self, system: System, target_node: Node, insert_resource: bool = False):
        """
        Add a system to the target node.
        :param system: System object
        :param target_node: Node object,  must be within the OSHConnect instance
        :param insert_resource: Whether to insert the system into the target node's server, default is False
        :return:
        """
        if target_node in self._nodes:
            target_node.add_new_system(system)
            if insert_resource:
                system.insert_self()
            self._systems.append(system)
            return

    def create_and_insert_system(self, system_opts: dict, target_node: Node):
        """
        Create a system on the target node.
        :param system_opts: System object parameters
        :param target_node: Node object, must be within the OSHConnect instance
        :return: the created system
        """
        if target_node in self._nodes:
            new_system = System(**system_opts)
            self.add_system(new_system, target_node, insert_resource=True)
            return new_system

    def remove_system(self, system_id: str):
        pass



#   ==============================================================================
#   Copyright (c) 2024 Botts Innovative Research, Inc.
#   Date:  2024/6/26
#   Author:  Ian Patterson
#   Contact Email:  ian@botts-inc.com
#   ==============================================================================

from typing import Callable

from pydantic import BaseModel


class VisualizationDataLayer(BaseModel):
    """
    Represents the data portion of a particular visualization as well as its state information.
    """

    name: str
    description: str
    datasource_id: list[str]
    visible: bool
    timestamp: float
    _get_timestamp: Callable
    on_left_click: Callable
    on_right_click: Callable
    on_hover: Callable
    _id: str
    _type: str
    _filter: dict
    _datasources_to_fn: Callable
    _no_datasources_fn: Callable


class Styling:

    def __init__(self):
        pass



#   ==============================================================================
#   Copyright (c) 2024 Botts Innovative Research, Inc.
#   Date:  2024/7/1
#   Author:  Ian Patterson
#   Contact Email:  ian@botts-inc.com
#   ==============================================================================
import websockets
from consys4py.comm.mqtt import MQTTCommClient
from consys4py.datamodels.commands import CommandJSON
from consys4py.datamodels.control_streams import ControlStreamJSONSchema

from oshconnect.osh_connect_datamodels import System


class ControlSchema:
    schema: dict = None


class ControlStream:
    name: str = None
    _parent_systems: System = None
    _strategy: str = "mqtt"
    _resource_endpoint = None
    # _auth: str = None
    _websocket: websockets.WebSocketServerProtocol = None
    _schema: ControlStreamJSONSchema = None
    _mqtt_client: MQTTCommClient = None

    def __init__(self, parent_system: System, resource_endpoint: str, name=None, strategy="mqtt"):
        self._parent_systems = parent_system
        self.name = name
        self._strategy = strategy
        self._resource_endpoint = resource_endpoint

    def set_schema(self, schema: ControlStreamJSONSchema):
        self._schema = schema

    def connect(self):
        pass

    def subscribe(self):
        if self._strategy == "mqtt" and self._mqtt_client is not None:
            self._mqtt_client.subscribe(f'{self._resource_endpoint}/commands')
        elif self._strategy == "mqtt" and self._mqtt_client is None:
            raise ValueError("No MQTT Client found.")
        elif self._strategy == "websocket":
            pass

    def publish(self, payload: CommandJSON):
        if self._strategy == "mqtt" and self._mqtt_client is not None:
            self._mqtt_client.publish(f'{self._resource_endpoint}/status', payload=payload, qos=1)
        elif self._strategy == "mqtt" and self._mqtt_client is None:
            raise ValueError("No MQTT Client found.")
        elif self._strategy == "websocket":
            pass

    def disconnect(self):
        pass

    def unsubscribe(self):
        self._mqtt_client.unsubscribe(f'{self._resource_endpoint}/commands')


class Command:
    pass
